from django.conf.urls import url, patterns
from django.conf import settings
from studiogdo.views import cached_javascript_catalog
from studiogdo.pip.views import LoginView, AccueilView

login_url = getattr(settings, 'LOGIN_URL')
if login_url.startswith("/"):
    login_url = login_url[1:]
urlpatterns = patterns('',
                       url(r'^%s$' % login_url, view=LoginView.as_view(), name="login"),
                       url(r'^$', view=AccueilView.as_view(), name="accueil"),
                       )


urlpatterns += patterns('',
                        url(r'^jsi18n/$', cached_javascript_catalog, name="javascript_catalog"),
                        )



# -*- coding: utf-8 -*-

import json

from django.core.exceptions import SuspiciousOperation
from django.http.response import HttpResponse
from django.test import TestCase
from mock import patch

from path import compose_path
from studiogdo.skel.renderer import HTMLRenderer, PathIterator


class TestPath(TestCase):
    def test_compose(self):
        path = compose_path('', '')
        self.assertEqual(path, '.')
        path = compose_path('', '.')
        self.assertEqual(path, '.')
        path = compose_path('a', '')
        self.assertEqual(path, 'a')
        path = compose_path('', 'b')
        self.assertEqual(path, 'b')
        path = compose_path('a', 'b')
        self.assertEqual(path, 'a/b')
        path = compose_path('a/b', 'c')
        self.assertEqual(path, 'a/b/c')
        path = compose_path('a', 'b/c')
        self.assertEqual(path, 'a/b/c')
        path = compose_path('a/b', 'c/d')
        self.assertEqual(path, 'a/b/c/d')
        path = compose_path('/', '/', '/')
        self.assertEqual(path, '/')
        path = compose_path('/a', '/b')
        self.assertEqual(path, '/b')
        path = compose_path('/', 'a')
        self.assertEqual(path, '/a')

    def test_compose_with_dot(self):
        path = compose_path('a', '.')
        self.assertEqual(path, 'a')
        path = compose_path('.', 'b')
        self.assertEqual(path, 'b')
        path = compose_path('a/./b', 'c/./d')
        self.assertEqual(path, 'a/b/c/d')
        path = compose_path('/', '.')
        self.assertEqual(path, '/')

    def test_compose_with_parent(self):
        path = compose_path('a/..', 'd/..')
        self.assertEqual(path, '.')
        path = compose_path('a/b/..', 'c/d/..')
        self.assertEqual(path, 'a/c')
        path = compose_path('a/b/..', 'd/..')
        self.assertEqual(path, 'a')
        path = compose_path('a/..', 'd')
        self.assertEqual(path, 'd')
        with self.assertRaises(SuspiciousOperation):
            compose_path('../a')
        with self.assertRaises(SuspiciousOperation):
            compose_path('a', '../..')
        path = compose_path('a/b/c', '../..')
        self.assertEqual(path, 'a')


class TestExtractPathes(TestCase):
    @patch('studiogdo.api.StudiogdoApi')
    def test_simple(self, mock_studiogdo):
        renderer = HTMLRenderer(mock_studiogdo, debug=True)

        skel = '<div debug="pass">test</div>'
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, set([]))

        skel = '<div data-cond="condition">test</div>'
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"condition"})

        skel = '<div data-cond="!condition">test</div>'
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"condition"})

        skel = '<div data-cond=".">test</div>'
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"."})

        skel = '<div data-cond="!">test</div>'
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"."})

        skel = '<div data-value="prop">test</div>'
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"prop"})

        skel = """
        <div data-path='path' data-value='prop'>
            <span data-value='prop'></span>
        </div>
        """
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"path", "prop", "path/prop"})

        skel = """
        <div data-path='path' data-value='prop'>
            <span data-path='comp' data-value='prop'></span>
        </div>
        """
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"path", "prop", "path/prop", "path/comp"})

        skel = """
        <div data-path='path' data-value='prop'>
            <span data-value='comp/prop'></span>
        </div>
        """
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"path", "prop", "path/comp/prop"})

        skel = """
        <div data-path='path' data-value='prop'>
            <span data-value='prop1'></span>
            <span data-value='prop2'></span>
        </div>
        """
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"path", "prop", "path/prop1", "path/prop2"})

        skel = """
        <select data-label='Nom'>
            <option data-path='Fixation' data-value='Id' data-label='Nom' />
        </select>
        """
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"Nom", PathIterator("Fixation", {"Id", "Nom"})})

        skel = """
        <select data-path="path" data-label='Nom'>
            <option data-path='Fixation' data-value='Id' data-label='Nom' />
        </select>
        """
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"path", "Nom", PathIterator("path/Fixation", {"Id", "Nom"})})

        skel = """
        <select data-cond="!path" data-label='Nom'>
            <option data-path='Fixation' data-value='Id' data-label='Nom' />
        </select>
        """
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"path", "Nom", PathIterator("Fixation", {"Id", "Nom"})})

        skel = """
        <table>
            <thead><tr data-path='Fixation'>
                <th data-value='Id' />
                <th data-value='Nom' />
            </tr></thead>
        </table>
        """
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {PathIterator("Fixation", {"Id", "Nom"})})

        skel = """
        <table data-path='.'>
            <thead><tr data-path='Fixation'>
                <th data-value='Id' />
                <th data-value='Nom' />
            </tr></thead>
        </table>
        """
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {".", PathIterator("Fixation", {"Id", "Nom"})})

        skel = """
        <table data-path='Support'>
            <thead><tr data-path='Fixation'>
                <th data-value='Id' />
                <th data-value='Nom' />
            </tr></thead>
        </table>
        """
        pathes = renderer.extract_pathes(skel)
        self.assertEqual(pathes, {"Support", PathIterator("Support/Fixation", {"Id", "Nom"})})

        with open('studiogdo/skel/examples/ex1.skel', 'r') as content_file:
            skel = content_file.read()
        pathes = renderer.extract_pathes(skel)
        inp = {".", "Adresse", "Adresse/Adresse1", "Adresse/Adresse2", "Adresse/CodePostal", "Adresse/Ville", "Adresse/Pays"}
        self.assertEqual(pathes, inp)

        with open('studiogdo/skel/examples/ex2.skel', 'r') as content_file:
            skel = content_file.read()
        pathes = renderer.extract_pathes(skel)
        ths = {"DateCommande", "SocieteFact", "NomFact", "DateLivraison"}
        tds = {"Id", "Transporteur", "prixFactTotalTTCAffiche", "SuiviTransport"}
        self.assertEqual(pathes, {".", PathIterator("CommandeValide", ths.union(tds))})

        for i in range(0, 0):
            with open('studiogdo/skel/examples/ex3.skel', 'r') as content_file:
                skel = content_file.read()
            pathes = renderer.extract_pathes(skel)
            ths = {"DateLivraison", "SocieteFact", "PrixFactHT", "prixFactTotalHT", "prixFactTotalTTCAffiche"}
            tds = {".^", "Id", "ReferenceClient", "DateCommande", "Status", "Delai", "NomFact", "PrenomFact", "ModePaiement", "TimePaiement"}
            self.assertEqual(pathes, {PathIterator(".", ths.union(tds))})


class TestRenderer(TestCase):
    @patch('studiogdo.api.StudiogdoApi')
    def test_renderer_cond(self, mock_studiogdo):
        renderer = HTMLRenderer(mock_studiogdo, debug=True)

        skel = '<div debug="pass">test</div>'
        res = renderer.render('/', skel)
        self.assertRegexpMatches(res, 'debug="pass">test</div>')
        self.assertFalse(mock_studiogdo.called)

        values = [{"data-path": "/", "Id": 0}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.return_value.content = js
        skel = '<div data-cond="condition">test</div>'
        res = renderer.render('/', skel)
        self.assertEqual(res.strip(), "")
        self.assertTrue(mock_studiogdo.get_list.called)

        values = [{"data-path": "/", "condition": "test", "Id": 0}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.return_value.content = js
        skel = '<div data-cond="!condition">test</div>'
        res = renderer.render('/', skel)
        self.assertEqual(res.strip(), "")
        self.assertTrue(mock_studiogdo.get_list.called)

        values = [{"data-path": "/", "condition": "test", "Id": 0}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.return_value.content = js
        skel = '<div data-cond="condition">test</div>'
        res = renderer.render('/', skel)
        self.assertRegexpMatches(res, 'debug="pass">test</div>')
        self.assertTrue(mock_studiogdo.get_list.called)

    @patch('studiogdo.api.StudiogdoApi')
    def test_renderer_container(self, mock_studiogdo):
        renderer = HTMLRenderer(mock_studiogdo, debug=True)

        values = [{"data-path": "/", ".": "/", "prop": "test", "Id": 0}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.return_value.content = js
        skel = "<span data-value='prop'></span>"
        res = renderer.render('/', skel)
        self.assertRegexpMatches(res, 'test')

        skel = "<span data-path='.'>test</span>"
        res = renderer.render('/', skel)
        self.assertRegexpMatches(res, 'test')

        skel = "<span data-path='!'>test</span>"
        res = renderer.render('/', skel)
        self.assertEquals(res, '')

    @patch('studiogdo.api.StudiogdoApi')
    def test_renderer_container_and_child(self, mock_studiogdo):
        renderer = HTMLRenderer(mock_studiogdo, debug=True)

        values = [{"data-path": "/", "path": "test1", "prop": "test3", "path/prop": "test2", "Id": 0}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.return_value.content = js
        skel = """<div data-path='path' data-value='prop'>
            <span data-value='prop'></span>
        </div>"""
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res.strip(), "test3test3")

        values = [{"data-path": "/", "path": "test1", "prop": "test3", "path/prop": "test2", "Id": 0}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.return_value.content = js
        skel = """<div data-path='path' data-value='prop'>
            <span data-value='path/prop'></span>
        </div>"""
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res.strip(), "test2test3")
        self.assertTrue(mock_studiogdo.get_list.called)

    @patch('studiogdo.api.StudiogdoApi')
    def test_renderer_container_not(self, mock_studiogdo):
        renderer = HTMLRenderer(mock_studiogdo, debug=True)

        values = [{"data-path": "/", "path": "test1", "prop": "test3", "path/prop": "test2", "Id": 0}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.return_value.content = js
        skel = """<div data-cond='!path'>
            test
        </div>"""
        res = renderer.render('/', skel)
        self.assertTrue(mock_studiogdo.get_list.called)
        self.assertRegexpMatches(res, '')

        values = [{"data-path": "/", "Id": 0}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.return_value.content = js
        mock_studiogdo.post_facet.return_value.content = ""
        skel = """<div data-cond='!path'>
            test
        </div>"""
        res = renderer.render('/', skel)
        self.assertTrue(mock_studiogdo.get_list.called)
        self.assertRegexpMatches(res.strip(), 'test')

    @patch('studiogdo.api.StudiogdoApi')
    def test_renderer_final(self, mock_studiogdo):
        renderer = HTMLRenderer(mock_studiogdo, debug=True)

        values = [{"data-path": "/", "prop": "aBc", "/prop": "aBc", "Id": 0}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.return_value.content = js

        skel = "<progress data-value='prop'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, 'aBc')

        skel = "<progress data-value='/prop'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, 'aBc')

        skel = "<progress data-value='prop%su'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, 'ABC')

        skel = "<progress data-value='prop%sl'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, 'abc')

        skel = "<progress data-value='prop%sc'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, 'Abc')

        values = [{"data-path": "/", "prop": "020", "Id": 0}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.return_value.content = js

        skel = "<progress data-value='prop'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, '020')

        skel = "<progress data-value='prop%i'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, '20')

        skel = "<progress data-value='prop%i*100'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, '2000')

        skel = "<progress data-value='prop%i/20'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, '1')

        skel = "<progress data-value='prop%i/100'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, '0')

        skel = "<progress data-value='prop%i/100###'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, '0.000')

        skel = "<progress data-value='prop%i/100,##'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, '0,00')

        skel = u"<progress data-value='prop%i/100,##€'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, u'0,00€')

        skel = u"<progress data-value='prop%i/100,## €'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, u'0,00 €')

        skel = u"<progress data-value='prop%i/100,##%'></progress>"
        res = renderer.render('/', skel, mode="text")
        self.assertEqual(res, u'0,00%')

    @patch('studiogdo.api.StudiogdoApi')
    def test_renderer_post(self, mock_studiogdo):
        renderer = HTMLRenderer(mock_studiogdo, debug=True)

        values = [{"data-path": "/", "prop": "50", "Id": 0}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.return_value.content = js
        skel = """<input data-value='prop'/>
            <input data-value='prop%s' name="s_"/>
            <input data-value='prop%i'/>
        """
        res = renderer.render('/', skel)
        self.assertRegexpMatches(res, '_L3Byb3A=')

    @patch('studiogdo.api.StudiogdoApi')
    def test_renderer_select(self, mock_studiogdo):
        """
        http://localhost:8080/morassuti/html/facet.gdo?p=/Services(baches)&f=python&m={%22data-path%22:%20%22Fixation%22,%20%22data-value%22:%20[%22Nom%22]}        """
        """
        {"data-path":"/Fixation","data-value":[{"data-path":"/Services(baches)/Fixation(1)","Nom":"Sans"},{"data-path":"/Services(baches)/Fixation(2)","Nom":"Colson/serflex blanc longueur 20cm"},{"data-path":"/Services(baches)/Fixation(3)","Nom":"Colson/serflex noir longueur 20cm"},{"data-path":"/Services(baches)/Fixation(4)","Nom":"Sandow/tendeur 1 crochets métal 20 cm"},{"data-path":"/Services(baches)/Fixation(5)","Nom":"Sandow/tendeur 1 crochets métal 40 cm"},{"data-path":"/Services(baches)/Fixation(6)","Nom":"Sandow/tendeur 2 crochets métal 20 cm"},{"data-path":"/Services(baches)/Fixation(7)","Nom":"Sandow/tendeur 2 crochets métal 40 cm"},{"data-path":"/Services(baches)/Fixation(8)","Nom":"Drisse polypro diam. 4 mm"},{"data-path":"/Services(baches)/Fixation(9)","Nom":"Drisse élastique diam. 8 mm"}]}    """
        renderer = HTMLRenderer(mock_studiogdo, debug=True)

        js1 = json.dumps({"data-path": "/", "data-value": [{"data-path": "/", "Nom": "Sans", "Id": 0}]})
        js2 = json.dumps({"data-path": "/Fixation", "Nom": "test", "data-value": [{"data-path": "/Fixation(0)", "Nom": "Sans", "Id": 0}, {"data-path": "/Fixation(1)", "Nom": "Colson/serflex blanc longueur 20cm", "Id": 1}, {"data-path": "/Fixation(2)", "Nom": "Colson/serflex noir longueur 20cm", "Id": 2}]})
        mock_studiogdo.get_list.side_effect = [HttpResponse(js1), HttpResponse(js2)]
        skel = """<select data-value='Nom'>
            <option data-path='Fixation' data-value='Nom' />
            </select>
        """
        res = renderer.render('/', skel)
        self.assertRegexpMatches(res, 'Colson/serflex blanc')
        self.assertRegexpMatches(res, 'Colson/serflex noir')
        self.assertRegexpMatches(res, 'selected="selected" value="Sans"')

        js1 = json.dumps({"data-path": "/", "data-value": [{"data-path": "/", "Nom": "Sans", "Id": "0"}]})
        js2 = json.dumps({"data-path": "/Fixation", "Nom": "test", "data-value": [{"data-path": "/Fixation(0)", "Nom": "Sans", "Id": "0"}, {"data-path": "/Fixation(1)", "Nom": "Colson/serflex blanc longueur 20cm", "Id": "1"}, {"data-path": "/Fixation(2)", "Nom": "Colson/serflex noir longueur 20cm", "Id": "2"}]})
        mock_studiogdo.get_list.side_effect = [HttpResponse(js1), HttpResponse(js2)]
        skel = """<select data-label='Nom'>
            <option data-path='Fixation' data-value='Id' data-label='Nom' />
            </select>
        """
        res = renderer.render('/', skel)
        self.assertRegexpMatches(res, 'selected="selected" value="0"')

        js1 = json.dumps({"data-path": "/", "data-value": [{"data-path": "/", "Nom": "Sans", "Id": "0"}]})
        js2 = json.dumps({"data-path": "/Fixation", "Nom": "test", "data-value": [{"data-path": "/Fixation(0)", "Nom": "Sans", "Id": "0"}, {"data-path": "/Fixation(1)", "Nom": "Colson/serflex blanc longueur 20cm", "Id": "1"}, {"data-path": "/Fixation(2)", "Nom": "Colson/serflex noir longueur 20cm", "Id": "2"}]})
        mock_studiogdo.get_list.side_effect = [HttpResponse(js1), HttpResponse(js2)]
        skel = """<select data-value='Nom'>
            <option value=''></option>
            <option value='1'></option>
            </select>
        """
        res = renderer.render('/', skel)
        self.assertRegexpMatches(res, 'option data-apath="Lw==" data-path-debug="/" value=""')

    @patch('studiogdo.api.StudiogdoApi')
    def test_renderer_ul(self, mock_studiogdo):
        renderer = HTMLRenderer(mock_studiogdo, debug=True)

        js1 = json.dumps({"data-path": "/", "data-value": [{"data-path": "/", "Nom": "Sans", "Id": "0"}]})
        js2 = json.dumps({"data-path": "/Fixation", "Nom": "test", "data-value": [{"data-path": "/Fixation(0)", "Nom": "Sans", "Id": "0"}, {"data-path": "/Fixation(1)", "Nom": "Colson/serflex blanc longueur 20cm", "Id": "1"}, {"data-path": "/Fixation(2)", "Nom": "Colson/serflex noir longueur 20cm", "Id": "2"}]})
        mock_studiogdo.get_list.side_effect = [HttpResponse(js1), HttpResponse(js2)]
        skel = """<ul data-cond='Nom'>
            <div data-value='Nom'></div>
            <li data-path='Fixation' data-value='Nom' />
            </ul>
        """
        res = renderer.render('/', skel)
        self.assertRegexpMatches(res, "Sans</span></li>")
        self.assertRegexpMatches(res, "Colson/serflex noir longueur 20cm</span></li></ul>")

        values = [{"data-path": "/", "prop": "test", "Id": "0"}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.side_effect = [HttpResponse(js)]
        skel = """<ul><li>
            <p data-value="prop"></p>
        </li></ul>
        """
        res = renderer.render('/', skel)
        self.assertRegexpMatches(res, "test")

    @patch('studiogdo.api.StudiogdoApi')
    def test_renderer_table(self, mock_studiogdo):
        renderer = HTMLRenderer(mock_studiogdo, debug=True)

        js1 = json.dumps({"data-path": "/", "data-value": [{"Support": "Sans", "Id": "0", "data-path": "/Support"}]})
        js2 = json.dumps({"data-path": "/Support", "data-value": [{"Nom": "Sans", "Id": "0", "data-path": "Fixation(1)"}, {"Nom": "Colson/serflex blanc longueur 20cm", "Id": "1", "data-path": "Fixation(2)"}, {"Nom": "Colson/serflex noir longueur 20cm", "Id": "2", "data-path": "Fixation(3)"}]})
        mock_studiogdo.get_list.side_effect = [HttpResponse(js1), HttpResponse(js2)]
        skel = """<table data-path='Support'>
            <thead><tr data-path='Fixation'>
                <th data-value='Id' />
                <th data-value='Nom' />
            </tr></thead>
            </table>
        """

        res = renderer.render('/', skel)
        self.assertRegexpMatches(res, '<tbody>')
        self.assertRegexpMatches(res, '<td>0</td><td>Sans</td>')
        self.assertRegexpMatches(res, '<td>1</td><td>Colson/serflex blanc longueur 20cm</td>')

        js1 = json.dumps({"data-path": "/", "data-value": [{"Support": "Sans", "Id": "0", "data-path": "/Support"}]})
        js2 = json.dumps({"data-path": "/Support", "data-value": [{"Nom": "Sans", "Id": "0", "data-path": "Fixation(1)"}, {"Nom": "Colson/serflex blanc longueur 20cm", "Id": "1", "data-path": "Fixation(2)"}, {"Nom": "Colson/serflex noir longueur 20cm", "Id": "2", "data-path": "Fixation(3)"}]})
        mock_studiogdo.get_list.side_effect = [HttpResponse(js1), HttpResponse(js2)]
        skel = """<table>
            <caption data-value='Support'>
            </caption>
            <thead><tr data-path='Fixation'>
                <th data-value='Id' />
                <th data-value='Nom' />
            </tr></thead>
            </table>
        """

        res = renderer.render('/', skel)

    @patch('studiogdo.api.StudiogdoApi')
    def test_renderer_data(self, mock_studiogdo):
        renderer = HTMLRenderer(mock_studiogdo, debug=True)

        values = [{"data-path": "/", "prop": "50", "Id": 0}]
        js = json.dumps({"data-path": "/", "data-value": values})
        mock_studiogdo.get_list.return_value.content = js
        skel = """<input data-prop-test1='prop' test1='10'/>
            <input data-prop-test2='prop1' test2='10'/>
            <input data-prop+test3='prop' test3='10 30'/>
            <input data-text-test4='<span data-value="prop"/>' test4='10'/>
            <input data-text-test5='<span data-value="prop1"/>' test5='10'/>
            <input data-text+test6='<span data-value="prop"/>' test6='10 30'/>
            <input data-html='<span data-value="prop"/>' test7='10 30'/>
        """
        res = renderer.render('/', skel)
        self.assertRegexpMatches(res, 'test1="50"')
        self.assertRegexpMatches(res, 'test2="10"')
        self.assertRegexpMatches(res, 'test3="10 30 50"')
        self.assertRegexpMatches(res, 'test4="50"')
        self.assertRegexpMatches(res, 'test5="10"')
        self.assertRegexpMatches(res, 'test7="10 30"')



# -*- coding: utf-8 -*-
import re
from collections import OrderedDict


def path(exp, skel, context=None, template_name=None):
    context = context or {}
    reg = re.compile(r'%s' % (exp,))
    if isinstance(skel, basestring):
        skel = [skel]
    return (reg, {"skel": skel, "context": context, "template_name": template_name})


class Patterns(object):
    def __init__(self, *kwargs):
        self._regexes = OrderedDict()
        for elt in kwargs:
            self._regexes[elt[0]] = elt[1]

    def __getitem__(self, name):
        # on rajoute dans le skel_dict les variables du regexp du path
        for regex, path_dict in self._regexes.items():
            m = regex.match(name)
            if m:
                path_group = m.groupdict()
                path_dict_return = path_dict.copy()
                path_dict_return.update({"path_group": path_group})
                return path_dict_return
        raise KeyError('Key %s does not match any regex' % name)

    def iterkeys(self):
        return self._regexes.iterkeys()

    def iteritems(self):
        return self._regexes.iteritems()

    def keys(self):
        return [reg.pattern for reg in self._regexes.iterkeys()]


patterns = Patterns



# -*- coding: utf-8 -*-

import re
import sys
import itertools
import json
import time
from base64 import b64encode as encode
from pprint import PrettyPrinter

import ipdb  # not really pythonic but veryusefull...
from bs4 import BeautifulSoup, NavigableString, Tag
from django.conf import settings

from path import compose_path


COND_REGEXP = re.compile(r"^(?P<value>[^!=<>]*)(?P<oper>(!|[!=<>]=?)?)(?P<slot>[^!=<>]*)$")

STRING_FORMAT = re.compile(r"^s(?P<value>(\[\])?)(?P<filter>[ulcw])?$")
INTEGER_FORMAT = re.compile(r"^i((\*(?P<mult>(\d)*))|(\/(?P<div>(\d)*)))?(?P<sep>[.,])?(?P<dec>\#*)(?P<unit>[^?]*)(?P<nothing>\??)$")
DATE_FORMAT = re.compile(r"^dt_(?P<exp>.*)$")

input_tags = ["input", "textarea"]
final_tags = ["img", "meter", "progress"]
container_tags = ["a", "abbr", "address", "article", "aside", "button", "blockquote",
                  "caption", "div", "fieldset", "figcaption", "footer", "form", "legend",
                  "h1", "h2", "h3", "h4", "h5", "h6", "header", "hgroup", "label", "nav", "p",
                  "pre", "span", "section", "tbody", "tr", "td", "thead", "tfoot",
]
list_tags = ["ul", "ol"]

iterator_tags = ["tr", "li", "option", "section"]


class SoupWrapper(object):
    def __init__(self):
        self.soup = []  # last document parsed on top

    def is_input(self, elt):
        """ Return True if the element is an input element """
        return elt.name in input_tags

    def is_container(self, elt):
        return elt.name in container_tags

    def get_children(self, elt):
        for child in (child for child in elt.children if isinstance(child, Tag)):
            yield child

    def get_immutable_children(self, elt):
        return [x for x in elt.children if isinstance(x, Tag)]

    def _insert_elt(self, elt, parent=None, before=None):
        if parent and before:
            raise Exception("cannot insert element with parent %s and sibling %s" % (parent, before))

        if before:
            before.insert_before(elt)
        elif parent:
            parent.append(elt)

    def append_new_tag(self, tag, parent=None, before=None):
        elt = self.soup[-1].new_tag(tag)
        self._insert_elt(elt, parent, before)
        return elt

    def clone_tag(self, elt, parent=None, before=None):

        # string case
        if isinstance(elt, NavigableString):
            string = type(elt)(elt)
            self._insert_elt(string, parent, before)
            return string

        # clone the element
        clone = Tag(None, elt.builder, elt.name, elt.namespace, elt.nsprefix)
        self._insert_elt(clone, parent, before)

        # work around bug where there is no builder set
        # https://bugs.launchpad.net/beautifulsoup/+bug/1307471
        clone.attrs = dict(elt.attrs)
        for attr in ('can_be_empty_element', 'hidden'):
            setattr(clone, attr, getattr(elt, attr))
        for child in elt.contents:
            self.clone_tag(child, parent=clone)
        return clone


class HTMLParser(object):
    def get_data_path(self, elt, attr='data-path'):
        """ Return path and True if operator not was found """
        path = elt.get(attr, ".")
        if path == ".":
            return ""

        if path:
            m = COND_REGEXP.match(path)
            if m:
                slot = m.group('slot') if m.group('oper') else m.group('value')
                if attr == 'data-path' and m.group('oper') == '!':
                    if self.strict:
                        raise Exception("not condition %s in path %s" % (path, elt))
                    elt["data-cond"] = path
                    del elt["data-path"]
                path = slot
        return path

    def get_all_pathes(self, elt, container='', exclude={}):
        """ Return all pathes used from a simple element """

        # search path used in attributes
        pathes = set()
        for k, v in elt.attrs.iteritems():
            if k in exclude:
                continue

            if v:

                # search from data-path
                if (k == "data-path" or k == "data-cond"):
                    path = self.get_data_path(elt, k)
                    pathes.add(compose_path(path, container))

                # search from data-value
                elif k in ["data-value", "data-label"]:
                    pourcent = v.find('%')
                    v = v if pourcent == -1 else v[:pourcent]
                    pathes.add(v)

                elif len(k) > 9 and k.startswith("data-prop"):
                    pourcent = v.find('%')
                    v = v if pourcent == -1 else v[:pourcent]
                    pathes.add(v)

                elif len(k) > 9 and k.startswith("data-text") or k.startswith("data-html"):
                    soup = BeautifulSoup(v, "html5")
                    p = self.get_all_pathes_recursiv(soup.body, container=container)
                    pathes.update(p)

        return pathes

    def get_all_pathes_recursiv(self, *args, **kwargs):
        """ Return all pathes used from an element and its children"""

        pathes = set()
        container = kwargs.get("container", "")
        for elt in (elt for elt in itertools.chain(*args) if isinstance(elt, Tag)):
            data_path = self.get_data_path(elt)

            # iterator tag with data-path are not in same scope
            if elt.name in iterator_tags:

                # specific case where tr is in tbody (data_path is then previous tr in thead)
                if elt.name == 'tr' and not data_path:
                    if elt.parent.name == "tbody":
                        thead = elt.parent.parent.thead
                        trs = thead.find_all("tr")
                        data_path = self.get_data_path(trs[0])

                # creates iterator path
                iterator_path = compose_path(container, data_path)
                props = self.get_all_pathes(elt, container=compose_path(container, data_path), exclude={'data-path'})
                new_props = self.get_all_pathes_recursiv(elt)  # pathes in iterator are relatives (no container)
                iterator_props = props.union(new_props)
                iterator = PathIterator(iterator_path, iterator_props)
                iterator.add_in(pathes)
                continue

            # add pathes from element
            for path in self.get_all_pathes(elt):
                pathes.add(compose_path(container, path))

            # search in children (data-path is used but not on first call)
            new_container = container
            if data_path and data_path != ".":
                new_container = compose_path(container, data_path)
                pathes.add(new_container)
            new_props = self.get_all_pathes_recursiv(self.get_children(elt), container=new_container)
            PathIterator.update(pathes, new_props)

        return pathes


    def get_value_path(self, elt, attr="data-value", path=None):
        """ Return value path and value format """
        path = path or elt.get(attr, "")
        index = path.find("%")

        # if no format defined, then use string format
        if index == -1:
            return path, "s"
        return path[:index], path[index + 1:]


class StudiogdoWrapper(object):
    def is_empty(self, stcl, slot):
        """ Test if a slot is empty """

        if not slot:
            raise Exception("is_empty : the property path should not be null")

        # get from stencil
        return stcl.get(slot) is None

    def get_value(self, stcl, path, _format="s", alert=True):
        """ Return property value and property path """

        if not path:
            raise Exception("get_value : the property path should not be null")

        if path.endswith("^"):
            prop = path[:-1]
            value_path = compose_path(stcl.get("data-path", ""), prop)
            value = encode(value_path)
            return self.format_value(value, _format)

        # if value is already stored
        value = stcl.get(path)
        if value is not None:
            return self.format_value(value, _format)
        composed = compose_path(stcl.get("data-path", ""), path)
        value = stcl.get(composed)
        if value is not None:
            return self.format_value(value, _format)

        if self.strict and alert:
            pp = PrettyPrinter()
            msg = 'get_value : %s not in stcl (%s)' % (path, pp.pformat(stcl))
            raise Exception(msg)

        return None

    def format_value(self, value, _format):

        found = STRING_FORMAT.search(_format)
        if found:
            if not value:
                return ""

            _filter = found.group('filter')
            if _filter:
                if _filter == 'u':
                    return value.upper()
                if _filter == 'l':
                    return value.lower()
                if _filter == 'c':
                    return value.capitalize()
            return value

        found = INTEGER_FORMAT.search(_format)
        if found:
            parsed = False

            if not value:
                nothing = found.group('nothing')
                if nothing:
                    return ""
                return 0

            mult = found.group('mult')
            if mult:
                parsed = True
                value = str(int(value) * int(mult))
            div = found.group('div')
            if div:
                parsed = True
                value = str(int(value) / int(div))
            dec = found.group('dec')
            if dec:
                parsed = True
                f = "%%.%sf" % (len(dec),)
                value = f % float(value)
            sep = found.group('sep')
            if sep:
                value = value.replace('.', sep)
            unit = found.group('unit')
            if unit:
                value = value + unit

            return value if parsed else str(int(value))

        found = DATE_FORMAT.search(_format)
        if found:
            if not value:
                return ""

            try:
                exp = found.group("exp")
                exp = exp.replace("dd", "%d")
                exp = exp.replace("MM", "%m")
                exp = exp.replace("yyyy", "%Y")
                exp = exp.replace("yyyy", "%Y")
                exp = exp.replace("hh", "%H")
                exp = exp.replace("mm", "%M")
                has_date = value.find('-') != -1
                has_time = value.find(':') != -1
                if has_date and has_time:
                    t = time.strptime(value, "%Y-%m-%d %H:%M")
                elif has_date:
                    t = time.strptime(value, "%Y-%m-%d")
                elif has_time:
                    t = time.strptime(value, "%H:%M:%S")
                return time.strftime(exp, t)
            except ValueError as e:
                return str(e)

        return value


class HTMLRenderer(SoupWrapper, HTMLParser, StudiogdoWrapper):
    TRANSACTION_ID = 0

    def __init__(self, api, mode="html", debug=settings.STUDIOGDO_DEBUG, strict=not settings.STUDIOGDO_DEBUG):
        self.api = api
        self.debug = debug
        self.lists = {}
        self.mode = mode
        self.strict = strict
        HTMLRenderer.TRANSACTION_ID = HTMLRenderer.TRANSACTION_ID + 1

        SoupWrapper.__init__(self)

    def extract_pathes(self, skel):
        soup = BeautifulSoup(skel, "html5")
        return self.get_all_pathes_recursiv(soup.body.children)


    def render(self, path, skel, mode=None, stcl=None):
        """
        Returns the skeleton completed.
        Accepted mode: text et html
        """
        soup = BeautifulSoup(skel, "html5")
        self.soup.append(soup)

        body = soup.body
        if not stcl:
            stcl = {"data-path": path, ".": path}
            stcl = self.get_stencil(stcl, path, body)
        self.expand_children(stcl, body)

        mode = mode or self.mode
        if mode == "html":
            result = u''.join((unicode(c) for c in body.contents))
        if mode == "text":
            result = body.get_text(strip=True)

        self.soup.pop()
        return result

    def expand(self, stcl, elt):
        """
        A stencil is a dict with at least a data-path entry
        """

        # debug infos
        if self.debug:
            elt["debug"] = "pass"
            debug = elt.get("data-debug")
            if settings.STUDIOGDO_DEBUG and debug == "stop":
                ipdb.set_trace()

        # check condition
        if not self.satisfy_condition(stcl, elt):
            elt.decompose()
            return
        self.expand_attributes(stcl, elt)

        # expand from type
        if self.is_container(elt):
            decomposed = self.expand_container(stcl, elt)
            if decomposed:
                return
        elif elt.name in final_tags:
            self.expand_final(stcl, elt)
        elif self.is_input(elt):
            self.expand_input(stcl, elt)
        elif elt.name in ["select"]:
            self.expand_select(stcl, elt)
        elif elt.name in list_tags:
            self.expand_list(stcl, elt)
        elif elt.name in ["table"]:
            self.expand_table(stcl, elt)

        self.remove_path(elt)

    def expand_children(self, stcl, elt):
        for child in self.get_children(elt):
            self.expand(stcl, child)

    def expand_container(self, stcl, container):
        """ Returns True if the container is decomposed """
        path = self.get_data_path(container)
        if path:

            # expand children
            empty = self.is_empty(stcl, path)
            if not empty:
                new_stencil = self.get_stencil(stcl, path, container)
                self.expand_children(new_stencil, container)
                self._set_final_value(stcl, container)
            else:
                container.decompose()
            return True
        else:
            self.expand_children(stcl, container)
            self._set_final_value(stcl, container)
        return False

    def expand_final(self, stcl, final):
        self._set_final_value(stcl, final)

    def _set_final_value(self, stcl, final):

        # expand to text value
        value_path, _format = self.get_value_path(final)
        if value_path:
            value = self.get_value(stcl, value_path, _format)
            if value:
                span = self.append_new_tag("span", final)
                span.append(value)
                path = self.get_data_path(final)
                self.set_data_path(span, compose_path(path, value_path))

    def expand_select(self, stcl, select):

        def set_selected(elt, value, by_value, multiple=False):
            """ set selected option """
            if value:
                for option in elt.find_all("option"):
                    if by_value:
                        selected = option["value"]
                    else:
                        selected = option.string
                    if value == selected:
                        option["selected"] = "selected"
                        if not multiple:
                            break

        # replace by generated options
        for child in self.get_immutable_children(select):

            # check condition
            if not self.satisfy_condition(stcl, child):
                child.decompose()
                continue
            self.expand_attributes(stcl, child)

            # expand child
            if child.name == "option":
                self._expand_option(stcl, select, child)

        # the selected stencil is a property
        value_path, _format = self.get_value_path(select)
        label_path = select.get("data-label")

        # get value for label (label may be different as value)
        path = self.get_data_path(select)
        p = compose_path(path, label_path if label_path else value_path)
        value = self.get_value(stcl, p, _format)

        # set post parameter
        name = select.get('name', "%s_" % _format if _format else "s_")
        select["name"] = name + encode(stcl.get("data-path", ""))

        # comparaison mode
        by_value = False if label_path else True

        # multiple selections
        if name.startswith("m_"):
            for v in value.split(':'):
                set_selected(select, v, by_value, multiple=True)

        # single selection
        else:
            set_selected(select, value, by_value)

    def _expand_option(self, stcl, select, option):

        path = self.get_data_path(option)
        value_path, value_format = self.get_value_path(option)
        label_path, label_format = self.get_value_path(option, attr="data-label")

        # the options are not generated from the stencil
        if not path:

            # if value path defined, then changes label and value
            if value_path:
                self._complete_option(stcl, value_path, value_format, label_path, label_format, option)

        # the options are generated
        else:

            # iterates over data-path of the option
            props = [value_path, label_path]
            stencils = self.get_list(stcl, path, props)
            try:
                for s in json.loads(stencils.content).get("data-value"):
                    opt = self.append_new_tag("option", before=option)
                    self._complete_option(s, value_path, value_format, label_path, label_format, opt)
            except Exception as e:
                traceback = sys.exc_info()[2]
                msg = "Error: %s[%s], %s" % (e, traceback.tb_lineno, stencils.content)
                raise Exception(msg)

            # remove option template
            option.decompose()

    def _complete_option(self, stcl, value_path, value_format, label_path, label_format, option):
        """ Complete option element """

        # expand attributes
        self.expand_attributes(stcl, option)

        # set label and value
        if label_path:
            option.string = self.get_value(stcl, label_path, label_path)
        else:
            option.string = self.get_value(stcl, value_path, value_format)
        option["value"] = self.get_value(stcl, value_path, value_format)

    def expand_list(self, stcl, elt):

        # expands all items
        for child in self.get_immutable_children(elt):

            # only take first li elements
            if child.name.lower() != "li":
                self.expand(stcl, child)
                continue

            # iterates over data-path of the list
            value_path, _ = self.get_value_path(child)
            if not value_path:
                self.expand_children(stcl, child)
                continue

            props = [value_path]
            path = self.get_data_path(child)
            stencils = self.get_list(stcl, path, props)
            try:
                for s in json.loads(stencils.content).get("data-value"):
                    # check condition
                    li = self.clone_tag(child, before=child)
                    self.expand_attributes(stcl, li)
                    self._set_final_value(s, li)
                    self.expand_children(s, li)
            except Exception as e:
                traceback = sys.exc_info()[2]
                msg = "Error: %s[%s], %s" % (e, traceback.tb_lineno, stencils.content)
                raise Exception(msg)

            child.decompose()

    def expand_table(self, stcl, table):
        thead = table.thead
        if thead:
            return self.expand_one_dimension_table(stcl, table)
        else:
            # pivot table as two dimension table;
            pass

    def expand_one_dimension_table(self, stcl, table):

        # the last tr in thead serve as a template for each row
        thead = table.thead
        trs = thead.find_all("tr")
        if not trs:
            self.expand_children(stcl, table)
            return

        # last tr in header must have a path attribute
        thr = trs[-1]
        pr = self.get_data_path(thr)

        # simple tr (not as iterator)
        if not pr:
            self.expand_children(stcl, table)
            return

        # complete others elements in table
        for elt in self.get_children(table):
            if not (elt.name.lower() == "thead" or elt.name.lower() == "tbody"):
                self.expand(stcl, elt)

        # tr in body may serve as template
        thh = thr.find_all("th")
        tbody = table.tbody
        tbr = tbody.tr if tbody else []
        tbd = tbr.find_all("td") if tbr else []

        # create tbody if doesn't exists
        if not tbody:
            tbody = self.append_new_tag("tbody", parent=table)

        # creates a row for each stencil
        props = self.get_all_pathes_recursiv(thh, tbd)
        stencils = self.get_list(stcl, pr, props)
        try:
            content = stencils.content
            if content:
                for s in json.loads(content).get("data-value"):
                    s[".."] = stcl
                    ntr = self.append_new_tag("tr", parent=tbody)
                    self.set_data_path(ntr, s.get("data-path"))
                    src = tbr if tbr else thr
                    for key in (key for key in src.attrs if not key == "data-path"):
                        ntr[key] = src.attrs[key]
                    self.expand_attributes(s, ntr)

                    # for each column declared in the header
                    tbdIndex = 0
                    for th in thh:
                        value_path, _format = self.get_value_path(th)

                        # create td from template
                        if self.td_from_header(th):
                            td = self.append_new_tag("td", parent=ntr)

                            # adds attributes and data class
                            # copyAndExpandAttributes(stclContext, s, th, td);
                            # String classPath = th.attr(CLASS_ATTRIBUTE);
                            # addClassToElement(stclContext, s, td, classPath);

                            # expands container associated for each row (only if
                            # data-path
                            # defined)
                            # others children (note relative to path) stay in header
                            for f in th.select("[data-path]"):
                                if self.is_container(f):
                                    ff = self.clone_tag(f, parent=td)
                                    self.expand(s, ff)

                        # td are defined in body (no content from stencil)
                        else:
                            if tbd and len(tbd) > tbdIndex:
                                td = self.clone_tag(tbd[tbdIndex], parent=ntr)
                                tbdIndex = tbdIndex + 1
                                # value_path, _format = self.get_value_path(td)
                                # expandAttributes(stclContext, s, td);
                            else:
                                td = self.append_new_tag("td", parent=ntr)

                        # adds cell content
                        if self.td_from_header(th):

                            # uses label for content (value can also be used)
                            if value_path and not value_path.startswith("!"):
                                value = self.get_value(s, value_path, _format)
                                td.string = unicode(value)
                        else:
                            td["data-path"] = "."
                            self.expand_container(s, td)

                            # add apath
                            # apath = compose(pwd, valuePath)
                            # setDataAPath(stclContext, td, apath)
        except ValueError as e:
            parent = table.parent
            span = self.append_new_tag("span", parent=parent)
            traceback = sys.exc_info()[2]
            span.append("Error: %s[%s], %s" % (e, traceback.tb_lineno, stencils.content))
            return

        # suppresses container with data-path from header (div, form)
        # (edit form should be first element)
        for th in thh:
            if self.td_from_header(th):
                for f in (f for f in th.select("[data-path]") if self.is_container(f)):
                    f.decompose()
            self.expand_children(stcl, th)

        # removes initial tbody content
        if tbr:
            tbr.decompose()

            # expands the tfoot if exists
            # Element tfoot = table.select("tfoot").first();
            # if (tfoot != null) {
            # completeChildren(stclContext, stcl, tfoot);
            # }

    def td_from_header(self, th):
        """ returns True is the tag has a value path or a label path """
        value_path = th.get("data-value")
        label_path = th.get("data-label")  # unusefull
        return value_path or label_path

    def expand_input(self, stcl, input):

        # expand to text value
        value_path, _format = self.get_value_path(input)
        if not value_path:
            return
        value = self.get_value(stcl, value_path, _format)

        # set value or select attribute
        tag = input.name.lower()
        if tag == "textarea":
            input["value"] = value
        elif tag == "input":
            _type = input.get('type')
            if _type == "checkbox":
                if value.lower() in ['true', '1', 'vrai', 'ok', 'o', 'oui']:
                    input["checked"] = "checked"
            elif _type == "radio":
                if value.lower() == input.get("value"):
                    input["checked"] = "checked"
            else:
                input["value"] = value

        # complete name with absolute path
        name = input.get('name')
        if not name:
            name = "%s_" % _format if _format else "s_"
        input["name"] = name + encode(compose_path(stcl.get("data-path"), value_path))

    def satisfy_condition(self, stcl, elt):
        cond = elt.get('data-cond')

        # check only if condition defined
        if not cond:
            return True

        # check comparision
        found = COND_REGEXP.search(cond)
        if found:
            value = found.group('value')
            oper = found.group('oper')
            slot = found.group('slot')

            # only slot name (value in value)
            if not slot:
                value = value if value else "."
                prop = self.get_value(stcl, value, _format="s", alert=False)
                if oper == "!":
                    return prop is None
                return prop is not None

            if oper == "!":
                prop = self.get_value(stcl, slot, _format="s", alert=False)
                return prop is None

            prop = self.get_value(stcl, slot, _format="s")
            return eval("'%s' %s '%s'" % (value, oper, prop))

        return False

    def expand_attributes(self, stcl, elt):
        """ Expand the element attributes """

        def set_attr(attr, value, sign):
            if value:
                if sign == "-":
                    elt[attr] = value
                else:  # sign == +
                    previous = elt.get(attr, "")
                    if type(previous) is list:
                        previous = ' '.join(previous)
                    elt[attr] = "%s %s" % (previous, value)

        # set path
        path = stcl.get("data-path")
        self.set_data_path(elt, path)

        # should be inlined here
        for key in [key for key in elt.attrs.keys() if len(key) > 9 and (key.startswith("data-prop") or key.startswith("data-text") or key.startswith("data-html"))]:
            sign = key[9]
            attr = key[10:]
            if key.startswith("data-prop"):
                value_path, _format = self.get_value_path(None, path=elt.attrs[key])
                value = self.get_value(stcl, value_path, _format)
            elif key.startswith("data-text"):
                value = self.render(path=None, skel=elt.attrs[key], mode="text", stcl=stcl)
            elif key.startswith("data-html"):
                value = self.render(path=None, skel=elt.attrs[key], mode="html", stcl=stcl)
            set_attr(attr, value, sign)
            del elt[key]

    def set_data_path(self, elt, path):
        """ Set element path attribute """

        if self.debug:
            elt['data-path-debug'] = path
        elt['data-apath'] = encode(path)

    def remove_path(self, elt):
        """ Remove all path attributes for security """
        del elt['data-path']
        del elt['data-value']
        del elt['data-label']
        del elt['data-cond']

    def get_list(self, stcl, path, props):
        data_path = stcl.get("data-path")
        key = "%s:%s:%s" % (data_path, path, props)

        # check not already called
        already = self.lists.get(key)
        if already:
            return already

        # store result
        stencils = self.api.get_list(data_path, path, props, tid=HTMLRenderer.TRANSACTION_ID)
        self.lists[key] = stencils
        return stencils

    def get_stencil(self, stcl, path, container):

        if self.debug:
            self.api.log.debug("get_stencil : from %s get %s in %s" % (stcl, path, container))

        # specific cases ("." may be not defined at root)
        if path == "." and stcl.get("."):
            return stcl

        # parent is defined only in case of included call
        if path == "..":
            return stcl.get("..")

        # get new stencil properties
        props = self.get_all_pathes_recursiv([container])
        if self.debug:
            self.api.log.debug("get_stencil : props are %s" % (props,))
        if props:
            values = self.get_list(stcl, path, props)
            try:
                content = values.content
                if content:
                    stencils = json.loads(content).get("data-value")
                    if self.debug:
                        self.api.log.debug("get_stencil : %s" % values.content)
                    return stencils[0]
                msg = 'get_stencil is empty'
                raise Exception(msg)
            except:
                msg = 'get_stencil : %s' % values.content
                raise Exception(msg)

        return stcl


class PathIterator:
    def __init__(self, path, props):
        self.path = path
        self.props = set(itertools.ifilter(lambda x: x != ".", props))

    def __str__(self):
        return self.path

    def __repr__(self):
        return '(' + self.path + ', [' + ', '.join(self.props) + '])'

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        if isinstance(other, PathIterator):
            return other.path == self.path and other.props == self.props
        return False

    @staticmethod
    def update(pathes, other):
        for o in other:
            if isinstance(o, PathIterator):
                o.add_in(pathes)
            else:
                pathes.add(o)

    def add_in(self, pathes):
        iterator = itertools.ifilter(lambda x: isinstance(x, PathIterator) and x.path == self.path, pathes)
        try:
            iterator.next().props.update(self.props)
        except StopIteration:
            pathes.add(self)



# -*- coding: utf-8 -*-

import re
import json
from base64 import b64decode

from django.views.generic import TemplateView
from django.utils.decorators import method_decorator
from django.template import Context
from django.template.loader import get_template_from_string
from django.views.decorators.csrf import csrf_exempt
from django.http.response import HttpResponse
from django.conf import settings
from django.utils import timezone
from django.views.decorators.http import last_modified
from django.views.i18n import javascript_catalog

from studiogdo.api import StudiogdoApi


class SkelView(TemplateView):
    """
    skel should be defined or get_skel is then called (as template_name and get_template_name).
    """

    @property
    def skel(self):
        if not hasattr(self, "_skel"):
            self._skel = self.get_skel()
        return self._skel

    def get_skel(self):
        return self.data.get('m')

    def get_template_names(self):
        return self.skel;

    @property
    def data(self):
        if not hasattr(self, "_data"):
            self._data = getattr(self.request, self.request.method)
        return self._data

    def get_entry(self, **kwargs):
        return kwargs['entry']

    def get_facet(self):
        return self.data.get('f')

    def get_facet_mode(self):
        if isinstance(self.skel, list):
            skel = self.skel[0]
        else:
            skel = self.skel
        return 'json' if skel.endswith("jskel") else 'dom5'


    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        resp = None
        if settings.STUDIOGDO_FACET == "django":
            entry = self.get_entry(**kwargs)
            facet = self.get_facet()

            # django facet entry is not done on java
            if entry == "facet" and facet == StudiogdoApi.DEFAULT_MODE:
                entry = "empty"

            # should be optimized to be called only if needed
            api = request.studiogdo_api
            response = api.send(request.method.lower(), '%s.gdo' % entry, self.data)

            # on error 412 return error message
            if response.status_code == 412:
                valid = re.compile(r"<b>message</b> <u>([^<]*)</u>")
                se = valid.search(response.content)
                return HttpResponse(status=412, content=se.groups(0))

            content = response.content
            if response.status_code == 200:

                # django facet are rendered by django
                if facet == StudiogdoApi.DEFAULT_MODE:
                    skel = self.get(request, *args, **kwargs)
                    ap = self.data.get('ap')
                    renderer = request.html_renderer
                    content = renderer.render(b64decode(ap) if ap else "/", skel.render().content.decode('utf-8'))

                # if it'a a facet, we add a new attribute to the response
                elif facet == 'json' and response.content:
                    response.json = json.loads(response.content)
                elif facet == 'dom5':
                    response.html = response.content

            # create django response
            content_type = response.headers[
                'content-type'] if 'content-type' in response.headers else 'text/html; charset=utf-8'
            resp = HttpResponse(content=content, content_type=content_type,
                                status=response.status_code, reason=response.reason)

            for header, value in response.headers.iteritems():
                # if header in ['content-length', 'content-disposition']:
                if header in ['content-disposition']:
                    resp[header] = value

        elif settings.STUDIOGDO_FACET == "java":

            # if the request has a facet, get it as a template
            if self.data.get('f', None) == StudiogdoApi.DEFAULT_MODE:
                skel = self.get(request, *args, **kwargs)
                self.data['m'] = skel.render().content
                # check if skel extension is jskel for json or skel to html facet
                self.data['f'] = self.get_facet_mode()

            api = request.studiogdo_api
            entry = self.get_entry(**kwargs)
            if request.FILES:
                content_type = None
                files = {} # self.data.dict();
                for k,v in request.FILES.iteritems():
                    files[v.name] = v
                    # files[k] = (v.name, v, v.content_type)

            else:
                content_type = 'application/x-www-form-urlencoded;charset=UTF-8'
                files = None
            facet_mode = self.data.get('f', "")
            settings_facet_mode = getattr(settings, "STUDIOGDO_FACET_MODE", None)
            if settings_facet_mode == "trans" and facet_mode == "dom5":
                self.data["f"] = settings_facet_mode
                facet_mode = settings_facet_mode
            response = api.send(request.method.lower(), '%s.gdo' % entry, data=self.data.dict(), content_type=content_type, files=files)

            # on error 412 return error message
            if response.status_code == 412:
                valid = re.compile(r"<b>message</b> <u>([^<]*)</u>")
                se = valid.search(response.content)
                return HttpResponse(status=412, content=se.groups(0))

            content = response.content
            if response.status_code == 200:

                # if it'a a facet, we add a new attribute to the response

                if facet_mode == 'json' and content:
                    response.json = json.loads(content)
                elif facet_mode in ('dom5', 'trans'):
                    if facet_mode == 'trans':
                        rederended_skel = get_template_from_string("{%% load i18n %%}\n%s" % content)
                        context = Context()
                        content = rederended_skel.render(context)
                    response.html = content

            # create django response
            content_type = response.headers['content-type'] if 'content-type' in response.headers else 'text/html; charset=utf-8'
            resp = HttpResponse(content=content, content_type=content_type, status=response.status_code, reason=response.reason)

            for header, value in response.headers.iteritems():
                # if header in ['content-length', 'content-disposition']:
                if header in ['content-disposition']:
                    resp[header] = value

        return resp


last_modified_date = timezone.now()
@last_modified(lambda req, **kw: last_modified_date)
def cached_javascript_catalog(request, domain='djangojs', packages=None):
    """override default javascript catalog view to cache it
    caches expires at server restart

    :param request:
    :param domain:
    :param packages:
    :return:
    """
    return javascript_catalog(request, domain, packages)



import base64
import logging
import json

import requests
from pprint import pformat
from django.conf import settings
from django.utils.translation import ugettext as _


default_facet_mode = getattr(settings, "STUDIOGDO_FACET_MODE", "dom5")


class StudiogdoApi(object):
    """
    Rest API to retrieve objects in a hierarchical back-end
    """

    # Studiogdo URL root
    DEFAULT_ROOT = "http://%s/%s" % (settings.STUDIOGDO_HOST,
                                     settings.STUDIOGDO_SERVLET)

    # Default facet mode
    DEFAULT_MODE = 'django'

    def __init__(self, root_address=DEFAULT_ROOT, options=None, cookies=None):
        self.root_address = root_address

        self.options = getattr(settings, 'STUDIOGDO_OPTIONS', {})
        if options is not None:
            self.options.update(options)
        self.params = {}

        # cookies from django
        self.cookies = cookies or {}

        # cookies from studiogdo (will be added to cookies on next call)
        self.studiogdo_cookies = {}

        self.log = logging.getLogger("studiogdo")

    def append_param(self, param, value):
        self.params[param] = value


    #
    # HTML FACET
    #
    # Path are encoded
    #

    def get_empty(self):
        """ Empty RPC entry to retrieve tomcat cookie. """
        return self.send('get', 'empty.gdo')

    def disconnect(self):
        """ Disconnect user connected."""
        return self.send('get', 'disconnect.gdo')

    def get_prop(self, path):
        """ Simple get RPC entry to retrieve a property value. """
        if path:
            self.append_param("ap", path)
        return self.send('get', 'prop.gdo')

    def post_empty(self, path=None, form_data=None):
        if path:
            self.append_param("ap", path)
        return self.send("post", "empty.gdo", form_data)

    def _post_facet(self, cmd, path, skeleton, facet="html5", form_data=None):
        if path:
            self.append_param("ap", path)
        self.append_param("m", skeleton)
        self.append_param("f", facet if facet else StudiogdoApi.DEFAULT_MODE)
        return self.send('post', "%s.gdo" % cmd, form_data)

    def post_facet(self, path, skeleton, facet="html5", form_data=None):
        return self._post_facet('facet', path, skeleton, facet, form_data)

    def post_facets(self, path, skeleton, facet="html5", form_data=None):
        return self._post_facet('facets', path, skeleton, facet, form_data)

    def apply_command(self, path, command, form_data=None):
        if path:
            self.append_param("ap", path)
        self.append_param("c", command)
        return self.send('post', "apply.gdo", form_data)

    #
    # PYTHON FACET
    #
    #   Path are not encoded
    #

    def get_list(self, path, slot, attrs=None, tid=None):

        # import ipdb; ipdb.set_trace()
        if settings.STUDIOGDO_DEBUG:
            self.log.debug("path:%s, slot:%s, attrs:%s", path, slot, attrs)

        if path:
            self.append_param("ap", path)
        if tid:
            self.append_param("tid", tid)
        _list = '[%s]' % ', '.join("\'%s\'" % a for a in attrs)
        self.append_param("m", '{"data-path":"%s", "data-value":%s}' % (slot, _list))
        self.append_param("f", "python")
        response = self.send('get', 'facet.gdo')
        if response.content.find("JsonParseException") != -1:
            raise Exception(response.content)
        return response

    #
    #   PYTHON INTERFACE
    #

    def send(self, method, url, data=None, content_type='application/x-www-form-urlencoded;charset=UTF-8', files=None):
        """ Return the studiogdo cookies, the XML result of a studiogdo call."""

        headers = {"content-type": content_type}

        request_kwargs = {'url': '%s/%s' % (self.root_address, url),
                          'headers': headers,
                          }

        # multipart
        if content_type is None:
            request_kwargs = {'url': '%s/%s' % (self.root_address, url)}

        # add studiogdo cookies
        if self.cookies:
            request_kwargs['cookies'] = self.cookies

        # set default options in parameters
        self.params.update(self.options)
        if "acceptNoStencil" not in self.params:
            self.append_param("acceptNoStencil", "true" if settings.STUDIOGDO_DEBUG else "false")
        if "country" not in self.params:
            country = getattr(settings, "STUDIOGDO_COUNTRY", None)
            if country:
                self.append_param("country", country)
        # set data in parameters
        data = data or {}
        self.params.update(data)
        data_name = "data" if method.lower() == "post" else "params"
        request_kwargs[data_name] = self.params
        if files:
            request_kwargs["files"] = files
        # request_kwargs["data"] = self.params
        # request_kwargs["params"] = self.params

        try:
            response = getattr(requests, method)(**request_kwargs)
        except:
            raise Exception("studiogdo service not available (url: %s)" % request_kwargs['url'])

        if not response.ok:
            try:
                reason = u"%s" % response.reason
            except UnicodeDecodeError:
                reason = u"%s - ISO-8859-1" % response.reason.decode("ISO-8859-1")
            error = u"studiogdo error (code: %s, url: %s) : %s" % (response.status_code, request_kwargs['url'], reason)
            params = pformat(self.params, indent=2)
            raise Exception("\n".join([error.encode("utf8"),params]))
        elif url == "apply.gdo" and default_facet_mode == "trans":
            content = json.loads(response.content)
            for k,v in content.iteritems():
                if isinstance(v, list):
                    content[k] = [_(vi) for vi in v]
                elif isinstance(v, basestring) and not v.isdigit():
                    content[k] = [_(vi) for vi in v]
            else:
                response._content = json.dumps(content)

        # reset params for next call
        self.params = {}

        # memorize the cookie for several studiogdo calls in same request
        if response.cookies:
            self.cookies = response.cookies
            self.studiogdo_cookies = response.cookies

        return response



import datetime
from django.contrib.auth import logout
from studiogdo.api import StudiogdoApi
from studiogdo.skel.renderer import HTMLRenderer


class APICookies(object):

    def process_request(self, request):
        """ Store tomcat cookie to preserve session """

        # create the Studiogdo API and HTML renderer
        api = StudiogdoApi(cookies=request.COOKIES)
        request.studiogdo_api = api
        request.html_renderer = HTMLRenderer(api)

        return None

    def process_response(self, request, response):
        """ Reset tomcat cookie to preserve session """

        # add studiogdo cookies (if process_request was done)
        if hasattr(request, "studiogdo_api"):
            for cookie in request.studiogdo_api.studiogdo_cookies:
                response.set_cookie(key=cookie.name, value=cookie.value,
                                    expires=cookie.expires, path=cookie.path,
                                    domain=cookie.domain, secure=cookie.secure)

        return response


class Timeout(object):

    def process_request(self, request):
        return None
        if request.user.is_authenticated():
            if 'lastRequest' in request.session:
                elapsedTime = datetime.datetime.now() - request.session['lastRequest']
                if elapsedTime.seconds > 15*60:
                    del request.session['lastRequest']
                    logout(request)

            request.session['lastRequest'] = datetime.datetime.now()
        else:
            if 'lastRequest' in request.session:
                del request.session['lastRequest']

        return None



from django.utils.module_loading import import_by_path
from django.conf import settings

from studiogdo.pip.admin import admin

pathpatterns = import_by_path("pip.admin.%s.pathpatterns" % admin.paths)
